"""
Association and interaction testing module for PANAMA
(1) association testing
(2) interaction testing
(3) parallelization using ipython
"""

import sys, os, string, cPickle, copy, pdb
import logging as LG
import scipy as SP
import numpy as N
import scipy.stats as st
import copy
import testing_core

try:
    from IPython.kernel import client
    ipy_version = 0.10
except ImportError:
    from IPython.parallel import Client as client
    ipy_version = 0.12

tc_clients = None

def interface(S, P, K, covs = None, I = None, model = 'LMM', return_fields= ['pv'], parallel = False,
	      add_mean = True, jobs = 1, file_directory=None, delta_opt_params=None, Ftest = False):

    """
    interface function for association testing
    S: SNPS [indiv X snps]
    Y: expr [indic X genes]
    covs: covariates 

    return_filed: fields to caculate and return
    file_directory: if provided data handling and result storing will be handled via file IO
    delta_opt_params: parameters for delta optimization 
    """

    #1. check dimensions
    [Nn, Np] = P.shape
    Ns       = S.shape[1]

    #check input arguments
    assert (S.shape[0]==Nn), 'dimension error'
    if K is not None:
        assert (K.shape[0]==Nn), 'dimension error'
        assert (K.shape[1]==Nn), 'dimension error'
    if I is not None:
        assert (I.shape[0]==Nn), 'dimension error'
    #covariates: 
    if covs is None:
	covs = SP.zeros([Nn,0])
        if add_mean:
            #add mean column ( we should generally do this )
            covs = SP.concatenate((covs,SP.ones([Nn,1])),axis=1)

    #figure out base path
    cwd=os.getcwd()
    base_path      = os.path.dirname(__file__)
    glob_base_path = os.path.join(cwd,base_path)
    paths = [glob_base_path]

    if file_directory is not None:
        if not os.path.exists(file_directory):
            print "file mode for testing specified but directory %s not existing; disableing" % file_directory

    #check whether to use file mode
    file_mode = ((file_directory is not None) and os.path.exists(file_directory))
    #setup file handling if switched on
    in_file = None
    in_file_path = None
    out_file = None
    out_file_base = None
    if file_mode:
        import h5py
        in_file = os.path.join(file_directory,'data.hdf5')
	in_file_path = os.path.abspath(in_file)
        out_file_base = os.path.join(file_directory,'result_')
        #store matrix objects
        _data = {'S':S,'P':P,'K':K,'I':I,'covs': covs}
        f = h5py.File(in_file,'w')
        for key in _data:
            if _data[key] is None:
                continue
            f.create_dataset(name=key,data=_data[key],chunks=True)
        f.close()
        del(f)

    if not parallel:
        RV = testing_core.wrapper(S, P, K, covs = covs, I = I, model = model,
				  path_append = paths,index_s = SP.arange(S.shape[1]),
				  index_p = SP.arange(P.shape[1]),
				  in_file = in_file_path,
				  delta_opt_params=delta_opt_params, Ftest = Ftest)        
    else:
	# setup engines and perform checks
	tc, eng_ids = start_engines(jobs)

	# split the data according to the number of jobs
        #note jobs is the max. number of jobs...
	Y_split = split_jobs(P, jobs)

	ipython_jobs = []
	for n in xrange(len(Y_split)):
	    i0 = Y_split[n][0]
	    i1 = Y_split[n][1]
	    P_n = Y_split[n][2]

            #arguments of job
            #convert all paths to absolute paths
            xargs=SP.array([S, P_n, K, covs, I, range(i0, i1), range(S.shape[1]),
			    model, return_fields, paths, out_file, in_file_path,
			    Ftest, delta_opt_params],dtype='object')
            #if in file_mode, overwrite:
            if file_mode:
                out_file = '%s%03d.pickle' % (out_file_base,n)
                if os.path.exists(out_file):
                    os.remove(out_file)
                xargs[0:4] = None
                xargs[-3]  = os.path.abspath(out_file)

	    if ipy_version == 0.10:
		job = client.MapTask(testing_core.wrapper, xargs.tolist())
		ipython_jobs.append(tc.run(job))
	    else:
		dview = eng_ids[:]
		lbview = tc.load_balanced_view()
		job = lbview.apply(testing_core.wrapper, *xargs.tolist())
		ipython_jobs.append(job)
	    
        #wait for job completion:
	if ipy_version == 0.10:
	    tc.barrier(ipython_jobs)
	else:
	    tc.wait(ipython_jobs)

	RV = None
	for n in xrange(len(Y_split)):
	    i0 = Y_split[n][0]
	    i1 = Y_split[n][1]

            #fecht results from job
	    if ipy_version == 0.10:
		_rv = tc.get_task_result(ipython_jobs[n])
	    else:
		_rv = ipython_jobs[n].get()
	    
            if isinstance(_rv,str):
		f = open(_rv,'rb')
                _rv = cPickle.load(f)
		f.close()
                _rv = testing_core.filter_dict(_rv,return_fields)

            #do we need to create result structures?
            if RV is None:
                RV = {}
                #get shape of all fields we need
                #append dimensions form result structure
                for key in _rv.keys():
                    val = _rv[key]
                    dim0 = [Ns,Np]
                    dim0.extend(list(val.shape[2::]))
                    RV[key] = SP.empty(dim0)
            #slot in result fileds
            for key in _rv.keys():
                RV[key][:,i0:i1] = _rv[key]
    return [RV[key] for key in return_fields]


#### internal helper functions ####


 
        

def split_jobs(Y, Njobs):
    #slit phenotype matrix into jobs
    #think about splitting snps also
    splits = []


    [N, Np] = Y.shape
    #maximal splitting range is one job per phenotype
    Njobs = min(Njobs,Np)

    #figure out phenotypes per job (down rounded)
    npj   = int(SP.floor(SP.double(Np)/Njobs))
    
    i0 = 0
    i1 = npj
    for n in xrange(Njobs):
        if n==(Njobs-1):
            #make sure last jobs spans all the rest.
            i1 = Np
        Y_ = Y[:,i0:i1]
	splits.append([i0, i1, Y_])
        #nex split
        i0 = i1
        i1 = i1 + npj
        
    return splits

def start_engines(jobs):
    """
    Starts the engines and performs some checks

    jobs: int
    """

    global tc_clients
    
    if ipy_version == 0.10:
	tc = client.TaskClient()
	mec = client.MultiEngineClient()
	eng_ids = mec.get_ids()
    else:
	if tc_clients == None:
	    tc = client()
	    eng_ids = tc.ids
	    tc_clients = tc
	else:
	    tc = tc_clients
	    eng_ids = tc.ids
	
    #reset engines
    #mec.reset()
    assert eng_ids > 0, "No engine started"	
    #assert eng_ids > jobs, "Not enough engines for the number of jobs required"

    return tc, eng_ids

                




import numpy
from distutils.core import setup
from distutils.extension import Extension
from Cython.Distutils import build_ext
import sys, os, pdb

compile_flags = ["-O3"]
# stupid mac is stupid
if sys.platform != "darwin" and sys.platform != "linux2":
    compile_flags.append("-march=native")
elif sys.platform == "darwin":
    os.environ['ARCHFLAGS'] = "-arch i386 -arch x86_64"

ext_modules = [Extension(
    name="lmm",
    language="c++", # grr, mac g++ -dynamic != gcc -dynamic
    sources=["lmm.pyx",
	     "c_lmm.cpp",
             "utils/Gamma.cpp",
             "utils/FisherF.cpp",
             "utils/Beta.cpp",
             "utils/MathFunctions.cpp",
             "utils/BrentC.cpp"
             ],
    include_dirs = [numpy.get_include()],  # .../site-packages/numpy/core/include
    extra_compile_args = compile_flags,
    )]

setup(
    name = 'lmm',
    cmdclass = {'build_ext': build_ext, 'inplace': True},
    ext_modules = ext_modules
    )



import numpy as np
import scipy as sp
import sys, pickle, pdb
import scipy.stats as st
import scipy.interpolate

def estimate(pv, m = None, verbose = False, lowmem = False, pi0 = None):
    """
    Estimates q-values from p-values

    Args
    =====

    m: number of tests. If not specified m = pv.size
    verbose: print verbose messages? (default False)
    lowmem: use memory-efficient in-place algorithm
    pi0: if None, it's estimated as suggested in Storey and Tibshirani, 2003. 
         For most GWAS this is not necessary, since pi0 is extremely likely to be
         1

    """
    
    assert(pv.min() >= 0 and pv.max() <= 1), "p-values should be between 0 and 1"

    original_shape = pv.shape
    pv = pv.ravel() # flattens the array in place, more efficient than flatten() 

    if m == None:
        m = float(len(pv))
    else:
        # the user has supplied an m
        m *= 1.0

    # if the number of hypotheses is small, just set pi0 to 1
    if len(pv) < 100 and pi0 == None:
        pi0 = 1.0
    elif pi0 != None:
        pi0 = pi0
    else:
        # evaluate pi0 for different lambdas
        pi0 = []
        lam = sp.arange(0, 0.90, 0.01)
        counts = sp.array([(pv > i).sum() for i in sp.arange(0, 0.9, 0.01)])
        
        for l in range(len(lam)):
            pi0.append(counts[l]/(m*(1-lam[l])))

        pi0 = sp.array(pi0)

        # fit natural cubic spline
        tck = sp.interpolate.splrep(lam, pi0, k = 3)
        pi0 = sp.interpolate.splev(lam[-1], tck)
        
        if pi0 > 1:
            if verbose:
                print("got pi0 > 1 (%.3f) while estimating qvalues, setting it to 1" % pi0)
            
            pi0 = 1.0

    assert(pi0 >= 0 and pi0 <= 1), "pi0 is not between 0 and 1: %f" % pi0


    if lowmem:
        # low memory version, only uses 1 pv and 1 qv matrices
        qv = sp.zeros((len(pv),))
        last_pv = pv.argmax()
        qv[last_pv] = (pi0*pv[last_pv]*m)/float(m)
        pv[last_pv] = -sp.inf
        prev_qv = last_pv
        for i in xrange(int(len(pv))-2, -1, -1):
            cur_max = pv.argmax()
            qv_i = (pi0*m*pv[cur_max]/float(i+1))
            pv[cur_max] = -sp.inf
            qv_i1 = prev_qv
            qv[cur_max] = min(qv_i, qv_i1)
            prev_qv = qv[cur_max]

    else:
        p_ordered = sp.argsort(pv)    
        pv = pv[p_ordered]
        qv = pi0 * m/len(pv) * pv
        qv[-1] = min(qv[-1],1.0)

        for i in xrange(len(pv)-2, -1, -1):
            qv[i] = min(pi0*m*pv[i]/(i+1.0), qv[i+1])
        
        # reorder qvalues
        qv_temp = qv.copy()
        qv = sp.zeros_like(qv)
        qv[p_ordered] = qv_temp

        # reshape qvalues
        qv = qv.reshape(original_shape)
        
    return qv



import numpy as np
import sys
sys.path.append('./../../')
import panama3
import pandas
import cPickle as pickle
from sklearn import metrics
import pylab as plt
# from panama.core import run

expr = pickle.load(open('/mnt/nicolo_storage/limmi/simulation/results/expr.pickle', 'r'))
snps = pickle.load(open('/mnt/nicolo_storage/limmi/simulation/results/snps.pickle', 'r'))
truth = pickle.load(open('/mnt/nicolo_storage/limmi/simulation/results/true_associations.pickle', 'r'))
cov = np.ones((expr.shape[0], 1))


conf = panama.core.ConfounderGPLVM(expr, snps, covariates=cov, population_structure=None, num_factors=5)
conf.fit()
qv, pv = conf.association_scan()
fpr, tpr, thresholds = metrics.roc_curve(truth.flatten(), -pv.flatten())
plt.figure()
plt.plot(fpr, tpr)

# qv1, pv1 = run.PANAMA(expr, snps, n_factors = 5, parallel=False)
# fpr, tpr, thresholds = metrics.roc_curve(truth.flatten(), -pv1.flatten())
# plt.plot(fpr, tpr)



import scipy as SP
import scipy.linalg as LA
import scipy.optimize as OPT
import scipy.stats as st
from pygp.linalg import *
import scipy.lib.lapack.flapack
import pdb

def nLLeval(ldelta,X,Y,K):
    """evaluate the negative LL of a LMM with kernel USU.T"""
    delta=SP.exp(ldelta)
    H = K + delta*SP.eye(X.shape[0])
    #the slow way: compute all terms needed
    L = jitChol(H)[0].T # lower triangular
    Linv = scipy.lib.lapack.flapack.dpotri(L)[0]
    Kinv = Linv.copy()
    SP.fill_diagonal(Linv, 0)
    Kinv += Linv.T
    n = X.shape[0]

    #1. get max likelihood weight beta
    pinv = SP.dot(SP.linalg.inv(SP.dot(SP.dot(X.T,Kinv),X)),X.T)
    y_   = SP.dot(Kinv,Y)
    beta = SP.dot(pinv,y_)
    #2. get max likelihood sigma
    yres = (Y-SP.dot(X,beta))
    R = SP.dot(SP.dot(yres.T,Kinv),yres)
    sigma2 = R/n

    nLL = 0.5*(-n*SP.log(2*SP.pi*sigma2) - 2*SP.log(L.diagonal()).sum() - 1./sigma2*R)
    return (-nLL);



def optdelta(X,Y,K,ldeltanull=None,numintervals=100,ldeltamin=-5.0,ldeltamax=5.0):
    """find the optimal delta"""
    if ldeltanull==None:
        nllgrid=SP.ones(numintervals+1)*SP.inf;
        #nllgridf=SP.ones(numintervals+1)*SP.inf;
        ldeltagrid=SP.arange(numintervals+1)/(numintervals*1.0)*(ldeltamax-ldeltamin)+ldeltamin;
        nllmin=SP.inf;
        ldeltamin=None;
        for i in SP.arange(numintervals+1):
            #nllgridf[i]=nLLevalf(ldeltagrid[i],UY,UX,S);
            nllgrid[i]=nLLeval(ldeltagrid[i],X,Y,K);
            if nllgrid[i]<nllmin:
                nllmin=nllgrid[i];
                ldeltamin=ldeltagrid[i];
        for i in SP.arange(numintervals-1)+1:
            ee=1E-8
            if ( ((nllgrid[i-1]-nllgrid[i])>ee) and (nllgrid[i+1]-nllgrid[i])>ee):
                #search within brent if needed
                ldeltaopt,nllopt,iter,funcalls = OPT.brent(nLLeval,(X,Y,K),(ldeltagrid[i-1],ldeltagrid[i],ldeltagrid[i+1]),full_output=True);
                if nllopt<nllmin:
                    nllmin=nllopt;
                    ldeltamin=ldeltaopt;
    else:
        ldeltamin=ldeltanull;
        #nllminf=nLLevalf(ldeltamin,UY,UX,S);
        nllmin=nLLeval(ldeltamin,X,Y,K);
        #assert SP.absolute(nllmin-nllminf)<1E-5, 'outch'
    return nllmin,ldeltamin;

def train(X,Y,K,covariates=None):
    """ compute all pvalues """
    n,s=X.shape;
    n_pheno=Y.shape[1];
    Xcovariate = SP.ones([X.shape[0],1])
    LL=SP.ones((n_pheno,s))*(-SP.inf);
    LL0=SP.ones((n_pheno))*(-SP.inf);
    pval=SP.ones((n_pheno,s))*(-SP.inf);
    for phen in SP.arange(n_pheno):
        nllnull,ldeltanull=optdelta(Xcovariate,Y[:,phen],K,ldeltanull=None,numintervals=100,ldeltamin=-5.0,ldeltamax=5.0);
        LL0[phen]=-nllnull;
        for snp in SP.arange(s):
            nll,ldelta=optdelta(SP.concatenate((X[:,snp:snp+1],Xcovariate),axis=1),Y[:,phen],K,ldeltanull=ldeltanull,numintervals=5,ldeltamin=ldeltanull-1.0,ldeltamax=ldeltanull+1.0);
            LL[phen,snp]=-nll;
            #calculate pvalues
            pval[phen,snp]=(st.chi2.sf(2.*(LL[phen,snp]-LL0[phen]), 1))
    return pval, LL0, LL

def eigenstratImputation(X, maxval=2.0):
    """
    impute SNPs as in
    'Principle components analysis corrects for stratification in genome-wide association studies'
    Price et al 2006 (Nature Genetics) page 908, Section 'Methhods', 'Inference of axes of variation'

    Input:
      X:        [n_indiv * n_SNP] matrix of SNPs, missing values are SP.nan
      maxval:   maximum value in X (default is 2 assuming 012 encoding)

    Output:
      X_ret:        [n_ind * n_SNP] matrix of imputed and standardized SNPs
    """
    n_i,n_s=X.shape
    X_ret=X.copy()
    for snp in SP.arange(n_s):
        i_nonan=X[:,snp]==X[:,snp]
        n_obs=SP.sum(i_nonan)
        snp_sum=SP.sum(X[i_nonan,snp])
        one_over_sqrt_pi=(1.+snp_sum)/(2.0+maxval*n_obs)
        one_over_sqrt_pi=1./SP.sqrt(one_over_sqrt_pi*(1.-one_over_sqrt_pi))
        snp_mean=(snp_sum*1.0)/(n_obs*1.0)
        X_ret[i_nonan,snp]-=snp_mean
        X_ret[i_nonan,snp]*=one_over_sqrt_pi
    X_ret[X!=X]=0.0
    return X_ret



import numpy as np
import scipy as sp
import GPy
import testing
import qvalue

def scaleK(K, verbose=False):
    """scale covariance K such that it explains unit variance"""
    c = sp.sum((sp.eye(len(K)) - (1.0 / len(K)) * sp.ones(K.shape)) * sp.array(K))
    scalar = (len(K) - 1) / c
    if verbose:
        print 'Kinship scaled by: %0.4f' % scalar
    K = scalar * K
    return K

def gs_cofficient(v1, v2):
    return np.dot(v2, v1) / np.dot(v1, v1)

def multiply(cofficient, v):
    return map((lambda x : x * cofficient), v)

def proj(v1, v2):
    return multiply(gs_cofficient(v1, v2) , v1)

def gs(X):
    Y = []
    for i in range(len(X)):
        temp_vec = X[i]
        for inY in Y :
            proj_vec = proj(inY, X[i])
            temp_vec = map(lambda x, y : x - y, temp_vec, proj_vec)
        Y.append(temp_vec)
    return Y

class ConfounderGPLVM(object):
    def __init__(self, expr, snps, covariates=None, population_structure=None, associations=True, interactions=False, num_factors=None, max_iterations=20, max_num_factors=30, FDR_assoc = 0.2):

        self.Y = expr
        self.S = snps
        self.N, self.D = expr.shape
        self.Q1 = self.S.shape[1]
        self.Y_means = self.Y.mean(axis=0)
        self.Y_stds = self.Y.std(axis=0)
        self.Y -= self.Y_means
        self.Y /= self.Y_stds
        self.S_centered = (self.S - self.S.mean(axis=0))/self.S.std(axis=0)
        self.candidate_associations = []

        self.max_num_factors = max_num_factors
        if num_factors is None:
            self.Q = self.Q_init()
            self.factors_how = 'detected'
        else:
            self.Q = num_factors
            self.factors_how = 'user-supplied'

        self.cov = covariates
        if covariates is None:
            self.num_cov = 0
        else:
            self.num_cov = covariates.shape[1]

        self.pop = population_structure
        kernel = self.kernel()

        initial_input_dim = self.Q+self.num_cov
        if self.pop is not None:
            initial_input_dim += 1

        self.model = GPy.models.GPLVM(self.Y, initial_input_dim, kernel=kernel)
        self.X = self.model.X[:, :self.Q]
        self.update_inputs()
        self.model.ensure_default_constraints()
        self.FDR_associations = FDR_assoc
        self.FDR_interactions = 0.2
        self.max_iterations = max_iterations
        self.panama_mode = associations
        self.limmi_mode = interactions

        print self


    def __str__(self):
        model_str = "Model: "

        if self.panama_mode:
            model_str += "associations (PANAMA) "
        if self.panama_mode and self.limmi_mode:
            model_str += "+"
        if self.limmi_mode:
            model_str += "interactions (LIMMI)"

        model_str += "\n%d samples, %d transcripts, %d SNPs, %d latent factors (%s)" % (self.N, self.D, self.Q1, self.Q, self.factors_how)
        model_str += "\nAssociation FDR: %f, Interaction FDR: %f" % (self.FDR_associations, self.FDR_interactions)
        model_str += "\nSNPs in covariance: %d " % len(self.candidate_associations)
        model_str += "\nMaximum training iterations: %d " % self.max_iterations
        if self.cov is not None:
            model_str += "\nAccounting for %d covariates" % self.cov.shape[1]
        if self.pop is not None:
            model_str += "\nAccounting for population structure"

        return model_str + "\n\n"

    def update_inputs(self):
        ncand = len(self.candidate_associations)
        kernel = self.kernel()
        self.model.kern = kernel

        # 1. Factors
        inputs = self.X

        # 2. Covariates
        if self.num_cov != 0:
            inputs = np.concatenate((inputs, self.cov), axis=1)

        # 3. Population structure
        if self.pop is not None:
            inputs = np.concatenate((inputs, np.zeros((self.N, 1))), axis=1)

        # 4. SNPs
        inputs = np.concatenate((inputs, self.S_centered[:, self.candidate_associations]), axis=1)

        self.model.X = inputs.copy()
        self.model.input_dim = self.model.X.shape[1]
        self.model._set_params(self.model._get_params())
        self.constrain_nonlatent()


    def kernel(self):
        num_candidates = len(self.candidate_associations)
        input_dim = self.Q + num_candidates + self.num_cov

        # 1. Factors
        kernel = GPy.kern.linear(self.Q)
        kernel.parts[0].name = 'confounders'

        # 2. Covariates
        if self.cov is not None:
            kernel = GPy.kern.kern.add(kernel, GPy.kern.linear(self.num_cov, ARD=True), tensor=True)
            kernel.parts[-1].name = 'covariates'

        # 3. Population structure
        if self.pop is not None:
            if self.pop is True:
                K = scaleK(np.dot(self.S, self.S.T))
            else:
                K = self.pop

            kernel = GPy.kern.kern.add(kernel, GPy.kern.fixed(1, K), tensor=True)
            kernel.parts[-1].name = 'popstruct'
            input_dim += 1

        # 4. SNPs
        if num_candidates != 0:
            kernel = GPy.kern.kern.add(kernel, GPy.kern.linear(num_candidates, ARD=True), tensor=True)
            kernel.parts[-1].name = 'genetics'

        kernel += GPy.kern.bias(input_dim)

        return kernel

    def kernel_testing(self, genetics=True, confounders=True):
        m = self.model.copy() # FIXME: this uses a fair bit of memory
        m.unconstrain('')
        if not genetics:
            m.constrain_fixed('genetics', 0.0)
        if not confounders:
            m.constrain_fixed('confounders', 0.0)
        m.constrain_fixed('X')
        m.ensure_default_constraints()
        m.optimize('bfgs', messages=0)
        return m.kern.K(m.X)

    def Q_init(self):
        """
        Returns the number of PCA dimensions that explain
        90% of the variance.

        """

        U, s, V = np.linalg.svd(self.Y.copy(), full_matrices = False)
        eigv = np.sort(s**2)[::-1]
        expl = np.cumsum(eigv)
        pca_dim = np.nonzero(expl/expl.max() > 0.90)[0].min()

        nf = min(self.max_num_factors, pca_dim)

        if nf == self.max_num_factors:
            print "WARNING: PCA detected a large number of latent dimensions (%d).\nWe reverted to the max_num_factors value (%d) to keep things tractable. \nIf you want, you can change the default max_num_factors value\n\n" % (pca_dim, self.max_num_factors)

        return nf

    def association_scan(self):
        print "Association scan... ",
        K = self.kernel_testing(genetics=False, confounders=True)
        pval = testing.interface(self.S_centered, self.Y, K, I = None,
                                 model='LMM', parallel=False, # TODO parallelize
                                 file_directory = None, jobs = 0)[0]
        print "[DONE]"
        # convert to qvalues
        qval = qvalue.estimate(pval)
        return qval, pval

    def interaction_scan(self):
        pass

    def get_latent(self, pca_rotation=True):
        pass

    def panama_step(self):
        X = self.model.X.copy()
        X = np.asarray(gs(X.T)).T
        X = np.linalg.qr(X)[0]
        X -= X.mean(axis = 0)
        X /= X.std(axis=0)
        K = self.kernel_testing(genetics=True, confounders=False)

        pv = testing.interface(self.S_centered, X, K, model = "LMM",
                               parallel = False, jobs = 0,
                               file_directory=None)[0] # TODO cleanup


        # Number of tests conducted
        num_tests = X.shape[1]*self.S.shape[1]
        qv = qvalue.estimate(pv, m=num_tests)

        # Set the qvalue of the current associations to 1
        qv[self.candidate_associations,:] = 1

        # Greedily construct addition set by adding the BEST (lowest qv) SNP for each factor
        # (if significant)
        new_candidates = []
        for i in xrange(qv.shape[1]):
            i_best = qv[:,i].argmin()
            qv_best = qv[:,i].min()
            # if significant, add it
            if qv_best<=self.FDR_associations:
                new_candidates.append(i_best)
                # and set the corrisponding qvalue to 1
                qv[i_best,:] = 1

        # Add candidates
        nc_old = len(self.candidate_associations)
        self.candidate_associations.extend(new_candidates)
        nc = len(self.candidate_associations)

        dl = nc-nc_old
        return dl

    def constrain_nonlatent(self):
        ind = range(self.Q, self.model.input_dim)
        fixed = ""
        for i in ind:
            fixed += "|%d" % i
        self.model._set_params(self.model._get_params())
        self.model.unconstrain('')
        self.model.constrain_fixed('X_\d+_(%s)$' % fixed)
        self.model.ensure_default_constraints()

    def limmi_step(self):
        K = np.eye(self.N) #TODO # self.Kpop

        # get latent factors and rotate to PCA
        X = self.getLatent(ard_reweight = True, pca_rotation = True, rotate_varimax = False)

        inter = interaction_scan.get_candidate_interactions(self.pheno, self.snps_raw, X, K,
                                                            interactions_iter = self.interactions_iter,
                                                            pv_max_stored = self.pv_max_stored,
                                                            topn_interactions = self.topn_interactions,
                                                            FDR_addition_interactions = self.FDR_addition_interactions,
                                                            candidate_interactions = self.candidate_interactions,
                                                            parallel = self.parallelize,
                                                            Njobs = self.Njobs)

        Nint_new, candidate_interactions, int_pv_qv = inter
        self.interactions_pv, self.interactions_qv = int_pv_qv
        # how to handle new interactions?
        if self.add_inter_fixed:
            # add fixed to covariance
            for i in xrange(len(candidate_interactions)):
                i_fact = candidate_interactions[i][0]
                i_snp  = candidate_interactions[i][1]
                # create local kernel matrix
                xx = (self.snps_raw[:,i_snp]*X[:,i_fact])[:,SP.newaxis]
                xx -= xx.mean(axis=0)
                xx /= xx.std(axis=0)
                # append to covariates
                self.covariates = SP.concatenate((self.covariates,xx),axis=1)
                self.nC = self.covariates.shape[1]
        else:
            # add to dynamic covariance relearning
            # note: this has issues as the tested interactions are on rotated factors...
            self.candidate_interactions=candidate_interactions

        return Nint_new

    def optimize(self):
        self.update_inputs()
        self.model.optimize('bfgs', messages=0)

    def fit(self):
        print("Training PANAMA with an association FDR of %.2f" % self.FDR_associations)

        if self.limmi_mode:
            print("Training PANAMA with an interaction FDR of %.2f" % self.FDR_interactions)

        it = 0

        train = True
        while train and it < self.max_iterations:
            print "Iteration %d" % it
            associations_added, interactions_added = 0, 0
            if it == 0:
                print "\t Optimizing factor model... ",
                self.optimize()
                self.X = self.model.X[:, :self.Q]
                print "[DONE]"
            else:
                if self.panama_mode:
                    print "\t running PANAMA step... ",
                    associations_added = self.panama_step()
                    print "[DONE]"
                    print "\t Added %d SNPs" % associations_added

                elif associations_added == 0 and self.limmi_mode:
                    print "\t running LIMMI step... ",
                    interactions_added = self.limmi_step()
                    print "[DONE]"

                print "\t Optimizing joint model... ",
                self.optimize()
                self.X = self.model.X[:, :self.Q]
                print "[DONE]"

                if associations_added == 0 and interactions_added == 0:
                    train = False

            it += 1
